/*

 vice.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>

#include "eth.h"
#include "crc8.h"

#ifndef MSG_NOSIGNAL
#define MSG_NOSIGNAL 0
#endif

#define SERVER_IP "127.0.0.1"
#define SERVER_PORT 64245

static int sock = -1;
static unsigned char ebufi[1024], ebufo[1024];
static unsigned int ebufip, ebufop, ebufil;
static struct hostent *hp;
static const char *i_addr;
static int i_port;
static int inited;

static const char *drivername(void) {
    return "VICE";
}

static int initialize(const char *addr, int port, FILE *lf, int lastfail) {
    int tr = 1;
    struct sockaddr_in eserver;

    i_addr = addr; i_port = port;
    inited = 0;

    if (!hp) hp = gethostbyname(addr ? addr : SERVER_IP);
    if (!hp) {
        if (lastfail != -1) fprintf(lf, "Hostname %s not found.\n", addr ? addr : SERVER_IP);
        return -1;
    }
    memset(&eserver, 0, sizeof(eserver));
    eserver.sin_family = AF_INET;
    memcpy(&eserver.sin_addr.s_addr, hp->h_addr, sizeof(eserver.sin_addr.s_addr));
    eserver.sin_port = htons(port ? port : SERVER_PORT);

    if (sock < 0) sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
        if (lastfail != -2) fprintf(lf, "Cannot open socket: %s(%d)\n", strerror(errno), errno);
        return -2;
    }
    if (connect(sock, (struct sockaddr *)&eserver, sizeof(eserver))<0) {
        if (lastfail != -3) fprintf(lf, "Cannot connect to %s: %s(%d)\n", addr ? addr : SERVER_IP, strerror(errno), errno);
        return -3;
    }
    setsockopt(sock, SOL_TCP, TCP_NODELAY, &tr, sizeof(tr));

    inited = 1;
    return 0;
}

static int getb(void) {
    unsigned char a;
    if (ebufip >= sizeof(ebufi) || ebufip >= ebufil) {
        int n = recv(sock, ebufi, sizeof(ebufi), MSG_NOSIGNAL);
        ebufip = ebufil = 0;
        if (n <= 0) return -EIO;
        ebufil = n;
    }
    a = ebufi[ebufip++];
    crc_add(a);
    return a;
}

static int sendb(unsigned char a) {
    if (ebufop >= sizeof(ebufo)) {
        if (send(sock, ebufo, ebufop, MSG_NOSIGNAL) < 0) {
            ebufop = 0;
            return -EIO;
        }
        ebufop = 0;
    }
    crc_add(a);
    ebufo[ebufop++] = a;
    return 0;
}

static int getbytes(unsigned short int bytes, unsigned char *data) {
    unsigned int i;
    while (ebufip + bytes > ebufil) {
        int n;
        ebufil -= ebufip;
        memcpy(data, ebufi + ebufip, ebufil);
        for (i = 0; i < ebufil; i++) crc_add(data[i]);
        data += ebufil; bytes -= ebufil;
        n = recv(sock, ebufi, sizeof(ebufi), MSG_NOSIGNAL);
        ebufip = ebufil = 0;
        if (n <= 0) return -EIO;
        ebufil = n;
    }
    memcpy(data, ebufi + ebufip, bytes);
    for (i = 0; i < bytes; i++) crc_add(data[i]);
    ebufip += bytes;
    return 0;
}

static int sendbytes(unsigned short int bytes, unsigned char *data) {
    unsigned int i;
    for (i = 0; i < bytes; i++) crc_add(data[i]);
    while (ebufop + bytes > sizeof(ebufo)) {
        memcpy(ebufo + ebufop, data, sizeof(ebufo) - ebufop);
        data += sizeof(ebufo) - ebufop;
        bytes -= sizeof(ebufo) - ebufop;
        ebufop = sizeof(ebufo);
        if (send(sock, ebufo, ebufop, MSG_NOSIGNAL) < 0) {
            ebufop = 0;
            return -EIO;
        }
        ebufop = 0;
    }
    memcpy(ebufo + ebufop, data, bytes);
    ebufop += bytes;
    return 0;
}

static void eshutdown(void) {
    if (sock >= 0) close(sock);
    sock = -1;
    inited = 0;
    return;
}

static int reinitialize(FILE *lf, int lastfail) {
    eshutdown();
    return initialize(i_addr, i_port, lf, lastfail);
}

static int flush(void) {
    if (!ebufop) return 0;
    if (send(sock, ebufo, ebufop, MSG_NOSIGNAL) < 0) {
        ebufop = 0;
        return -EIO;
    }
    ebufop = 0;
    return 0;
}

static void clean(void) {
    return;
}

static int waitb(void) {
    if (sock < 0 || !inited) return -ENODEV;
    return getb();
}

const struct io_ops_s vice_io_ops = {
    .drivername   = drivername,
    .initialize   = initialize,
    .reinitialize = reinitialize,
    .getb         = getb,
    .sendb        = sendb,
    .getbytes     = getbytes,
    .sendbytes    = sendbytes,
    .shutdown     = eshutdown,
    .flush        = flush,
    .wait         = waitb,
    .clean        = clean,
};
